@@ -1,45 +1,35 @@
 """
-inverter.py
-Density-to-potential inversion module
-
-Handles the primary functions
+Inverter.py
 """
-
-import numpy as np
+from warnings import warn
 from dataclasses import dataclass
+import numpy as np
 from opt_einsum import contract
 
-import psi4
-psi4.core.be_quiet()
-psi4.core.clean()
-psi4.set_options({"save_jk" : True})
-
-from .methods.wuyang import WuYang
 from .methods.zmp import ZMP
-from .methods.mrks import MRKS
-from .methods.oucarter import OC
+from .methods.wuyang import WuYang
 from .methods.pdeco import PDECO
+from .methods.oucarter import OC
+from .methods.mrks import MRKS
 from .methods.direct import Direct
-from .grid.grider import Grider
 
 @dataclass
-class data_bucket:
-    """
-    Data class for storing grid attributes. 
-    """
-    pass
+class V:
+    """Stores Potentials on AO"""
+    T : np.ndarray
 
+class E:
+    """Stores Energies"""
 
-class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
+class Inverter(Direct, ZMP, WuYang, PDECO, OC, MRKS):
     """
     Attributes:
     ----------
-    wfn : psi4.core.{RHF, UHF, RKS, UKS, Wavefunction, CCWavefuncion...}
-        Psi4 wavefunction object
-    mol : psi4.core.Molecule
-        Psi4 molecule object
-    basis : psi4.core.BasisSet
-        Psi4 basis set object
+
+    mol : Engine.molecule
+        Molecule class of engine used
+    basis : Engine.basis
+        Basis class of engine used
     basis_str : str
         Basis set
     nbf : int
@@ -53,13 +43,13 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
         1 -> Restricted
         2 -> Unrestricted
     Dt : List
-        List of psi4.core.Matrix for target density matrices (on AO).
+        List of np.ndarray for target density matrices (on AO).
     ct : List
-        List of psi4.core.Matrix for input occupied orbitals. This might not be correct for post-HartreeFock methods.
+        List of np.ndarray for input occupied orbitals. This might not be correct for post-HartreeFock methods.
     pbs_str: string
         name of Potential basis set
-    pbs : psi4.core.BasisSet
-        Potential basis set.
+    pbs : Engine.basis
+        Basis class for Potential basis set of the engine used. 
     npbs : int
         the length of pbs
     v_pbs : np.ndarray shape (npbs, ) for ref==1 and (2*npbs, ) for ref==2.
@@ -76,115 +66,135 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
         The three ao overlap matrix (ao, ao, pbs)
     S4  : np.ndarray
         The four ao overlap matrix, the size should be (ao, ao, ao, ao)
-    jk  : psi4.core.JK
-        Psi4 jk object. Built if wfn has no jk, otherwise use wfn.jk
-
+    jk  : Engine.jk
+        Engine jk object.
     T   : np.ndarray
         kinetic matrix on ao
     V   : np.ndarray
         external potential matrix on ao
     T_pbs: np.ndarray
         kinetic matrix on pbs. Useful for regularization.
-
     guide_potential_components: list of string
         guide potential components name
     va, vb: np.ndarray of shape (nbasis, nbasis)
         guide potential Fock matrix.
-    Methods:
-    --------
-
     """
-    def __init__(self, wfn, pbs="same", debug=False):
+
+    def __init__( self, engine='psi4' ):
+        self.eng_str = engine.lower()
+        if engine.lower() == 'psi4':
+            from .engines import Psi4Engine
+            self.eng = Psi4Engine()
+        elif engine.lower() == 'pyscf':
+            from .engines import PySCFEngine
+            self.eng = PySCFEngine()
+        else:
+            raise ValueError("Engine name is incorrect. The availiable engines are: {psi4, pyscf}")
+            
+    def __repr__( self ):
+        return "n2v.Inverter"
+
+    def set_system( self, molecule, basis, ref=1, pbs='same' , **kwargs):
         """
-        Handles Inversion
+        Stores relevant information and intitializes Engine
+
+        Parameters
+        ----------
+        molecule: Engine.molecule
+            Molecule object of selected engine
+        basis: str
+            Basis set of the main calculation
+        ref: int
+            reference for system. Restricted   -> 1
+                                  Unrestricted -> 2
+        pbs: str, default='same'
+            Basis set for the potential
+        **kwargs:
+            Optional Parameters for different Engiens 
+            Psi4 Engine:
+                wfn : psi4.core.{RHF, UHF, RKS, UKS, Wavefunction, CCWavefuncion...}
+                Psi4 wavefunction object
+            PySCF Engine:
+                None
+
+        """
+        # Communicate TO engine
+
+        self.eng.set_system(molecule, basis, ref, pbs, **kwargs)
+        self.ref = ref
+
+        self.nalpha = self.eng.nalpha
+        self.nbeta = self.eng.nbeta
+
+        # Initialize ecompasses everything the engine builds with basis set 
+        self.eng.initialize()
+        self.set_basis_matrices()
+
+        # Receive FROM engine
+        self.nbf  = self.eng.nbf
+        self.npbs = self.eng.npbs
+        self.v_pbs = np.zeros( (self.npbs) ) if self.ref == 1 \
+                                             else np.zeros( 2 * self.npbs )
+
+    @classmethod
+    def from_wfn( self, wfn, pbs='same' ):
+        """
+        Generates Inverter directly from wavefunction. 
         
         Parameters
         ----------
-        wfn : psi4.core.{RHF, UHF, RKS, UKS, Wavefunction, CCWavefuncion...}
-            Psi4 wavefunction object
-        pbs: str. default="same". If same, then the potential basis set (pbs)
-                 is the same as orbital basis set (i.e. ao). Notice that
-                 pbs is not needed for some methods
+        wfn: Psi4.Core.{RHF, RKS, ROHF, CCWavefunction, UHF, UKS, CUHF}
+            Wavefunction Object
+        Returns
+        -------
+        inv: n2v.Inverter
+            Inverter Object. 
         """
-        self.wfn       = wfn
-        self.pbs_str   = pbs
-        self.mol       = wfn.molecule()
-        self.basis     = wfn.basisset()
-        self.basis_str = wfn.basisset().name()
-        self.nbf       = wfn.basisset().nbf()
-        self.nalpha    = wfn.nalpha()
-        self.nbeta     = wfn.nbeta()
-        self.ref       = 1 if psi4.core.get_global_option("REFERENCE") == "RHF" or \
-                              psi4.core.get_global_option("REFERENCE") == "RKS" else 2
-        self.jk        = wfn.jk() if hasattr(wfn, "jk") == True else self.generate_jk()
-        self.Dt        = (np.array(wfn.Da_subset("AO")), np.array(wfn.Db_subset("AO")))
-        self.ct        = (np.array(wfn.Ca_subset("AO", "OCC")), np.array(wfn.Cb_subset("AO", "OCC")))
-        self.pbs       = self.basis if pbs == "same" \
-                                    else psi4.core.BasisSet.build( self.mol, key='BASIS', target=self.pbs_str)
-        self.npbs      = self.pbs.nbf()
-        self.v_pbs     = np.zeros( (self.npbs) ) if self.ref == 1 \
-                                                 else np.zeros( 2 * self.npbs )
-        self.generate_mints_matrices()
-        self.grid = data_bucket
-        self.cubic_grid = data_bucket
-        
-        self.J0 = None
-        self.S4 = None  # Entry to save the 4 overlap matrix.
+        from .engines import Psi4Engine
+        inv = self( engine='psi4' )
+        inv.eng = Psi4Engine()
+        ref = 1 if wfn.to_file()['boolean']['same_a_b_dens'] else 2
+        inv.set_system( wfn.molecule(), wfn.basisset().name(), pbs=pbs, ref=ref, wfn=wfn )
+        inv.Dt = [ np.array(wfn.Da()), np.array(wfn.Db()) ]
+        inv.ct = [ np.array(wfn.Ca_subset("AO", "OCC")), np.array(wfn.Cb_subset("AO", "OCC")) ]
+        inv.et = [ np.array(wfn.epsilon_a_subset("AO", "OCC")), np.array(wfn.epsilon_b_subset("AO", "OCC")) ]
+        inv.eng_str = 'psi4'
+        inv.eng.wfn = wfn
 
-    #------------->  Basics:
+        return inv
 
-    def generate_mints_matrices(self):
+    def set_basis_matrices( self ):
         """
-        Generates matrices that are methods of a mints object
+        Generate basis dependant matrices
         """
+        self.T  = self.eng.get_T()
+        self.V  = self.eng.get_V()
+        self.A  = self.eng.get_A()
+        self.S2 = self.eng.get_S()
+        self.S3      = self.eng.get_S3()
 
-        mints = psi4.core.MintsHelper( self.basis )
+        if self.eng.pbs_str != 'same':  
+            self.T_pbs  = self.eng.get_Tpbas()
 
-        #Overlap Matrices
-        self.S2 = np.array(mints.ao_overlap())
-        A = mints.ao_overlap()
-        A.power( -0.5, 1e-16 )
-        self.A = np.array(A)
-        self.S3 = np.squeeze(mints.ao_3coverlap(self.basis,self.basis,self.pbs))
+        self.S4 = None
 
-        #Core Matrices
-        self.T = np.array(mints.ao_kinetic()).copy()
-        self.V = np.array(mints.ao_potential()).copy()
-        self.T_pbs = np.array(mints.ao_kinetic(self.pbs, self.pbs)).copy()
-
-    def generate_jk(self, gen_K=False):
+    def compute_hartree( self, Cocc_a, Cocc_b ):
         """
-        Creates jk object for generation of Coulomb and Exchange matrices
-        1.0e9 B -> 1.0 GB
+        Computes Hartree Potential on AO basis set. 
+
+        Parameters
+        ----------
+        Cocc_a, Cocc_b: np.ndarray (nbf, nbf)
+            Occupied orbitals in ao basis
+
+        Returns
+        -------
+        J: List of np.ndarray
+            Hartree potential due to density from Cocc_a and Cocc_b
         """
-        jk = psi4.core.JK.build(self.basis)
-        memory = int(jk.memory_estimate() * 1.1)
-        jk.set_memory(int(memory)) 
-        jk.set_do_K(gen_K)
-        jk.initialize()
+        return self.eng.compute_hartree(Cocc_a, Cocc_b )
 
-        return jk
-
-    def form_jk(self, Cocc_a, Cocc_b):
-        """
-        Generates Coulomb and Exchange matrices from occupied orbitals
-        """
-        # if  self.jk.memory_estimate() > psi4.get_memory() * 0.8:
-        #     raise ValueError("Requested JK will take too more memory than default. \n \
-        #                       Increase it with psi4.set_memory(int( Many More Bytes )))!")
-        Cocc_a = psi4.core.Matrix.from_array(Cocc_a)
-        Cocc_b = psi4.core.Matrix.from_array(Cocc_b)
-        self.jk.C_left_add(Cocc_a)
-        self.jk.C_left_add(Cocc_b) 
-        self.jk.compute()
-        self.jk.C_clear()
-
-        J = (np.array(self.jk.J()[0]), np.array(self.jk.J()[1]))
-        K = []
-
-        return J, K
-
-    def diagonalize(self, matrix, ndocc):
+    def diagonalize( self, matrix, ndocc ):
         """
         Diagonalizes Fock Matrix
 
@@ -207,38 +217,98 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
             Eigenvalues
         """
 
-        A = self.A
-        Fp = A.dot(matrix).dot(A)
+        Fp = self.A.dot(matrix).dot(self.A)
         eigvecs, Cp = np.linalg.eigh(Fp)
-        C = A.dot(Cp)
+        C = self.A.dot(Cp)
         Cocc = C[:, :ndocc]
         D = contract('pi,qi->pq', Cocc, Cocc)
         return C, Cocc, D, eigvecs
 
-    #------------->  Inversion:
+    def diagonalize_with_potential_vFock(self, v=None):
+        """
+        Diagonalize Fock matrix with additional external potential
+        Stores values in object. 
 
-    def invert(self, method,
-                     guide_potential_components = ["fermi_amaldi"],
+        Parameters
+        ----------
+        v: np.ndarray
+            Additional external potential to be added to hamiltonian along with:
+            Kinetic_nm
+            External_nm
+            Guide_Potential_nm
+        """
+
+        if v is None:
+            fock_a = self.V + self.T + self.va
+        else:
+            if self.ref == 1:
+                fock_a = self.V + self.T + self.va + v
+            else:
+                valpha, vbeta = v
+                fock_a = self.V + self.T + self.va + valpha
+                fock_b = self.V + self.T + self.vb + vbeta
+
+
+        self.Ca, self.Coca, self.Da, self.eigvecs_a = self.diagonalize( fock_a, self.nalpha )
+
+        if self.ref == 1:
+            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.Ca.copy(), self.Coca.copy(), self.Da.copy(), self.eigvecs_a.copy()
+        else:
+            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.diagonalize( fock_b, self.nbeta )    
+
+    # Actual Methods
+    def generate_components(self, guide_components, **keywords):
+        """
+        Generates exact potential components to be added to
+        the Hamiltonian to aide in the inversion procedure. 
+        Parameters:
+        -----------
+        guide_potential_components: list
+            Components added as to guide inversion. 
+            Can be chosen from ["hartree", "fermi_amandi", "svwn"]
+        """
+
+        self.guide_components = guide_components
+        self.va = np.zeros( (self.nbf, self.nbf) )
+        self.vb = np.zeros( (self.nbf, self.nbf) )
+        self.J0 = self.compute_hartree(self.ct[0], self.ct[1])
+        N       = self.nalpha + self.nbeta
+
+        if self.eng_str == 'psi4':
+            J0_NO = self.eng.hartree_NO(self.Dt[0])
+            self.J0 = J0_NO if J0_NO is not None else self.J0
+
+        if guide_components == 'none':
+            warn("No guide potential was provided. Convergence may not be achieved")
+        elif guide_components == 'hartree':
+            self.va += self.J0[0] + self.J0[1]
+            self.vb += self.J0[0] + self.J0[1]
+        elif guide_components == 'fermi_amaldi':
+            v_fa = (1-1/N) * (self.J0[0] + self.J0[1])
+            self.va += v_fa
+            self.vb += v_fa
+        else:
+            raise ValueError("Guide component not recognized")
+
+    def invert(self, method, 
+                     guide_components = 'hartree',
                      opt_max_iter = 50,
                      **keywords):
         """
         Handler to all available inversion methods
-
         Parameters
         ----------
-
         method: str
             Method used to invert density. 
             Can be chosen from {wuyang, zmp, mrks, oc}. 
             See documentation below for each method. 
-        guide_potential_components: list, opt
+        guide_components: list, opt
             Components added as to guide inversion. 
             Can be chosen from {"fermi_amandi", "svwn"}
             Default: ["fermi_amaldi"]
         opt_max_iter: int, opt
             Maximum number of iterations inside the chosen inversion.
             Default: 50
-
         direct
         ------
         Direct inversion of a set of Kohn-Sham equations. 
@@ -276,7 +346,6 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                                 Notice that opt has lower priorities than opt_max_iter and gtol.
             return:
                 the result are stored in self.v_pbs
-
             
         zmp
         ---
@@ -317,8 +386,6 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                         vxc = self.on_grid_esp(Da=self.proto_density_a, Db=self.proto_density_b, grid=grid) for restricted;
                         vxc_up = self.on_grid_esp(Da=self.proto_density_a, Db=np.zeros_like(self.proto_density_a),
                                   grid=grid) for unrestricted;
-
-
         mRKS
         ----
         the modified Ryabinkin-Kohut-Staroverov method:
@@ -363,7 +430,6 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                     [2]: atol, [3]: atol1 for vxc_grid calculation.
             return:
                 The result will be stored in self.grid.vxc
-
         oc
         --
         Ou-Carter method
@@ -399,7 +465,6 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                     method name string that works for psi4. A separate psi4 calculation
                     would be performed.
                     wuyang
-
         pdeco
         -----
         the PDE-Constrained Optimization method:
@@ -429,15 +494,8 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                 the result are stored in self.v_pbs
         """
 
-        #Generate Guide Potential
-        if method.lower()=='mrks':
-            if guide_potential_components[0] != 'hartree' or len(guide_potential_components) != 1:
-                print("The guide potential is changed to v_hartree.")
-            self.generate_components(["hartree"])
-        elif method.lower() != 'direct':
-            self.generate_components(guide_potential_components)
+        self.generate_components(guide_components)
 
-        #Invert
         if method.lower() == "direct":
             return self.direct_inversion(**keywords)
         elif method.lower() == "wuyang":
@@ -445,82 +503,15 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
         elif method.lower() == "zmp":
             self.zmp(opt_max_iter, **keywords)
         elif method.lower() == "mrks":
+            if self.eng_str == 'pyscf':
+                raise ValueError("mRKS method not yet available with the PySCF engine. Try another method or another engine.")
             return self.mRKS(opt_max_iter, **keywords)
         elif method.lower() == 'oc':
+            if self.eng_str == 'pyscf':
+                raise ValueError("OuCarter method not yet available with the PySCF engine. Try another method or another engine.")
             return self.oucarter(opt_max_iter, **keywords)
         elif method.lower() == 'pdeco':
             return self.pdeco(opt_max_iter, **keywords)
         else:
             raise ValueError(f"Inversion method not available. Methods available: {['wuyang', 'zmp', 'mrks', 'oc', 'pdeco']}")
 
-    def generate_components(self, guide_potential_components):
-        """
-        Generates exact potential components to be added to
-        the Hamiltonian to aide in the inversion procedure. 
-
-        Parameters:
-        -----------
-        guide_potential_components: list
-            Components added as to guide inversion. 
-            Can be chosen from {"fermi_amandi", "svwn"}
-        """
-
-        self.va = np.zeros_like(self.T)
-        self.vb = np.zeros_like(self.T)
-
-        N = self.nalpha + self.nbeta
-
-        if self.J0 is None:
-            if type(self.wfn) == psi4.core.CCWavefunction:
-                nbf = self.nbf
-                C_NO = psi4.core.Matrix(nbf, nbf)
-                eigs_NO = psi4.core.Vector(nbf)
-                self.wfn.Da().diagonalize(C_NO, eigs_NO, psi4.core.DiagonalizeOrder.Descending)
-                occu = np.sqrt(eigs_NO.np)
-                New_Orb_a = occu * C_NO.np
-                assert np.allclose(New_Orb_a @ New_Orb_a.T, self.Dt[0])
-                if self.ref == 1:
-                    New_Orb_b = New_Orb_a
-                else:
-                    self.wfn.Db().diagonalize(C_NO, eigs_NO, psi4.core.DiagonalizeOrder.Descending)
-                    occu = np.sqrt(eigs_NO.np)
-                    New_Orb_b = occu * C_NO.np
-                self.J0 = self.form_jk(New_Orb_a, New_Orb_b)[0]
-            else:
-                self.J0, _ = self.form_jk(self.ct[0], self.ct[1])
-
-        if "fermi_amaldi" in guide_potential_components:
-            v_fa = (1-1/N) * (self.J0[0] + self.J0[1])
-
-            self.va += v_fa
-            self.vb += v_fa
-        elif "hartree" in guide_potential_components:
-            v_hartree = (self.J0[0] + self.J0[1])
-
-            self.va += v_hartree
-            self.vb += v_hartree
-        else:
-            raise ValueError("Hartee nor FA was included."
-                             "Convergence will likely not be achieved")
-
-        if "svwn" in guide_potential_components:
-            if "svwn" in guide_potential_components:
-                _, wfn_0 = psi4.energy( "svwn"+"/"+self.basis_str, molecule=self.mol , return_wfn = True)
-
-            if self.ref == 1:
-                ntarget = psi4.core.Matrix.from_array( [ self.Dt[0] + self.Dt[1] ] )
-                wfn_0.V_potential().set_D( [ntarget] )
-                vxc_a = psi4.core.Matrix( self.nbf, self.nbf )
-                wfn_0.V_potential().compute_V([vxc_a])
-                self.va += vxc_a.np
-                self.vb += vxc_a.np
-            elif self.ref == 2:
-                na_target = psi4.core.Matrix.from_array( self.Dt[0] )
-                nb_target = psi4.core.Matrix.from_array( self.Dt[1] )
-                wfn_0.V_potential().set_D( [na_target, nb_target] )
-                vxc_a = psi4.core.Matrix( self.nbf, self.nbf )
-                vxc_b = psi4.core.Matrix( self.nbf, self.nbf )
-                wfn_0.V_potential().compute_V([vxc_a, vxc_b])
-                self.va += vxc_a.np
-                self.vb += vxc_b.np
-        self.guide_potential_components = guide_potential_components
